# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import json
import logging
import subprocess
from functools import cached_property

from vasp_manager.utils import change_directory

logger = logging.getLogger(__name__)


class JobManager:
    """
    Handles job submission and status monitoring
    """

    def __init__(self, calc_path, ignore_personal_errors=True):
        """
        Args:
            calc_path (str): base directory of job
            ignore_personal_errors (bool): if True, ignore job submission errors
                if on personal computer
        """
        self.calc_path = calc_path
        self.ignore_personal_errors = ignore_personal_errors
        self._jobid = None
        self._job_complete = None

    @property
    def computing_config_dict(self):
        all_calcs_dir = self.calc_path.parent.parent
        fname = "computing_config.json"
        fpath = all_calcs_dir / "computing_config.json"
        if fpath.exists():
            with open(fpath) as fr:
                computing_config = json.load(fr)
        else:
            raise Exception(f"No {fname} found in path {all_calcs_dir.absolute()}")
        return computing_config

    @cached_property
    def computer(self):
        return self.computing_config_dict["computer"]

    @cached_property
    def user_id(self):
        return self.computing_config_dict[self.computer]["user_id"]

    @cached_property
    def mode(self):
        return self.calc_path.name

    @property
    def job_exists(self):
        jobid_path = self.calc_path / "jobid"
        if not jobid_path.exists():
            return False
        if jobid_path.stat().st_size == 0:
            return False

        with open(jobid_path) as fr:
            jobid = fr.read().strip()
        self.jobid = jobid
        return True

    @property
    def jobid(self):
        if not self.job_exists:
            raise Exception(f"jobid has not been set in {self.calc_path}")
        return self._jobid

    @jobid.setter
    def jobid(self, job_value):
        # some criteria to make sure jobid actually looks reasonable?
        # but I think SLURM will throw and error if sbatch fails
        try:
            jobid_int = int(job_value)
        except Exception as e:
            raise Exception(f"Tried to set jobid={job_value}\n{e}")
        self._jobid = jobid_int

    def submit_job(self):
        """
        Submits job, making sure to not make duplicate jobs
        """
        if self.job_exists:
            logger.info(f"{self.mode.upper()} Job already exists")
            return True

        if "personal" in self.computer:
            error_msg = f"Cannot submit {self.mode.upper()} job for on personal computer"
            error_msg += "\n\tIgnoring job submission..."
            logger.debug(error_msg)
            return True

        vaspq_path = self.calc_path / "vasp.q"
        if not vaspq_path.exists():
            logger.info(f"No vasp.q file in {self.calc_path}")
            # return False here instead of catching an exception
            # This enables job resubmission by letting the calling function
            # know that the calculation needs to be restarted
            return False

        submission_call = "sbatch vasp.q | awk '{ print $4 }'"
        with change_directory(self.calc_path):
            jobid = (
                subprocess.check_output(submission_call, shell=True)
                .decode("utf-8")
                .strip()
            )
            self.jobid = jobid
            with open("jobid", "w+") as fw:
                fw.write(jobid)
        logger.info(f"Submitted job {jobid}")
        return True

    @property
    def job_complete(self):
        if self._job_complete is None and self.job_exists:
            self._job_complete = self._check_job_complete()
        return self._job_complete

    def _check_job_complete(self):
        """Returns True if job done"""
        if self.computer == "personal":
            error_msg = "Cannot check job on personal computer"
            error_msg += "\n\tIgnoring job status check..."
            logger.debug(error_msg)
            # This enables job resubmission by letting the calling function
            # continue anyways
            return True

        check_queue_call = f"squeue -u {self.user_id}"
        queue_call = (
            subprocess.check_output(check_queue_call, shell=True)
            .decode("utf-8")
            .splitlines()
        )
        for line in queue_call:
            line = line.strip().split()
            if str(self.jobid) in line:
                return False
        return True



import shutil
import subprocess
import sys
from pathlib import Path

from vasp_manager import VaspManager
from vasp_manager.utils import change_directory

sys.path.append("../..")
from run_vasp_calculations import make_calculations_folder


def test_vmg_in_order(tmp_path):
    original_calculations_folder = Path("calculations")
    temp_calculation_folder = tmp_path
    if temp_calculation_folder.exists():
        shutil.rmtree(temp_calculation_folder)
    make_calculations_folder(calcs_path=temp_calculation_folder)

    for f in ["computing_config.json", "calc_config.json", "unzip_outputs.sh"]:
        shutil.copy(original_calculations_folder / f, temp_calculation_folder)

    calculation_types = [
        "rlx-coarse",
        "rlx",
        "static",
        "bulkmod",
        "elastic",
    ]
    material_paths = [
        p for p in sorted(list(temp_calculation_folder.glob("*"))) if p.is_dir()
    ]

    for i, calculation_type in enumerate(calculation_types):
        calculation_type_subset = calculation_types[: i + 1]
        vmg = VaspManager(
            calculation_types=calculation_type_subset,
            material_paths=material_paths,
            use_multiprocessing=True,
            to_rerun=True,
            to_submit=True,
        )
        results = vmg.run_calculations()
        if i != 0:
            previous_calc_types = calculation_type_subset[:i]
            # check results
            for material in results:
                for pc in previous_calc_types:
                    match pc:
                        case "rlx-coarse":
                            assert results[material][pc] == "done"
                        case "rlx":
                            assert isinstance(results[material][pc]["total_dV"], float)
                        case "static":
                            assert isinstance(
                                results[material][pc]["final_energy"], float
                            )
                        case "bulkmod":
                            assert isinstance(results[material][pc]["B"], float)
                        case _:
                            raise Exception
            # check summary
            summary = vmg.summary(as_string=False)
            for pc in previous_calc_types:
                assert summary["n_total"] == summary[pc]["n_finished"]

        for p in material_paths:
            mat_name = p.name
            orig_mode_path = original_calculations_folder / mat_name / calculation_type
            temp_mode_path = p / calculation_type
            shutil.rmtree(temp_mode_path)
            shutil.copytree(orig_mode_path, temp_mode_path)
        unzip_call = f"bash unzip_outputs.sh"
        with change_directory(temp_calculation_folder):
            subprocess.call(unzip_call, shell=True)

    vmg = VaspManager(
        calculation_types=calculation_types,
        material_paths=material_paths,
        to_rerun=True,
        to_submit=True,
    )
    results = vmg.run_calculations()
    for material in results:
        assert results[material]["rlx-coarse"] == "done"
        assert isinstance(results[material]["rlx"]["initial_spacegroup"], int)
        assert isinstance(results[material]["rlx"]["relaxed_spacegroup"], int)
        assert isinstance(results[material]["rlx"]["total_dV"], float)
        assert isinstance(results[material]["static"]["final_energy"], float)
        assert isinstance(results[material]["bulkmod"]["B"], float)
        assert isinstance(results[material]["elastic"]["B_VRH"], float)

    summary = vmg.summary(as_string=False)
    for calculation_type in calculation_types:
        assert summary["n_total"] == summary[calculation_type]["n_finished"]


def test_vmg_with_skipping(tmp_path):
    original_calculations_folder = Path("calculations")
    temp_calculation_folder = tmp_path
    if temp_calculation_folder.exists():
        shutil.rmtree(temp_calculation_folder)
    make_calculations_folder(calcs_path=temp_calculation_folder)

    for f in ["computing_config.json", "calc_config.json", "unzip_outputs.sh"]:
        shutil.copy(original_calculations_folder / f, temp_calculation_folder)

    calculation_types = [
        "rlx-coarse",
        "rlx",
        "static",
        "bulkmod",
        "elastic",
    ]
    material_paths = [
        p for p in sorted(list(temp_calculation_folder.glob("*"))) if p.is_dir()
    ]

    # skip static and bulkmod when copying
    completed_calc_types = ["rlx-coarse", "rlx", "elastic"]
    for calculation_type in completed_calc_types:
        for p in material_paths:
            mat_name = p.name
            orig_mode_path = original_calculations_folder / mat_name / calculation_type
            temp_mode_path = p / calculation_type
            shutil.copytree(orig_mode_path, temp_mode_path)

    unzip_call = f"bash unzip_outputs.sh"
    with change_directory(temp_calculation_folder):
        subprocess.call(unzip_call, shell=True)

    vmg = VaspManager(
        calculation_types=calculation_types,
        material_paths=material_paths,
        use_multiprocessing=True,
        to_rerun=True,
        to_submit=True,
    )
    results = vmg.run_calculations()
    for material in results:
        for calc_type in calculation_types:
            match calc_type:
                case "rlx-coarse":
                    assert results[material][calc_type] == "done"
                case "rlx":
                    assert isinstance(results[material][calc_type]["total_dV"], float)
                case "static" | "bulkmod":
                    assert results[material][calc_type] is None
                case "elastic":
                    assert isinstance(results[material][calc_type]["B_VRH"], float)
                case _:
                    raise Exception

    summary = vmg.summary(as_string=False)
    for pc in calculation_types:
        match pc:
            case "rlx-coarse" | "rlx" | "elastic":
                assert summary["n_total"] == summary[pc]["n_finished"]
            case "bulkmod" | "static":
                assert summary[pc]["n_finished"] == 0
            case _:
                raise Exception



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import logging
from pathlib import Path

import numpy as np
from pymatgen.analysis.eos import BirchMurnaghan
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun

logger = logging.getLogger(__name__)


class BulkmodAnalyzer:
    def __init__(self, calc_path=None, rounding_precision=3):
        """
        Args:
            calc_path (str): path to bulkmod calculation folder
            rounding_precision (int): precision to round calculated quantities
        """
        self._calc_path = Path(calc_path) if calc_path else calc_path
        self._rounding_precision = rounding_precision
        self._results = None

    @property
    def calc_path(self):
        if self._calc_path is not None:
            self.calc_path = self._calc_path
        return self._calc_path

    @calc_path.setter
    def calc_path(self, value):
        # make sure cij wasn't already defined
        if not value.exists():
            raise ValueError(f"Could not set calc_path to {value} as it does not exist")
        self._calc_path = value

    @property
    def rounding_precision(self):
        if self._rounding_precision is not None:
            self.rounding_precision = self._rounding_precision
        return self._rounding_precision

    @rounding_precision.setter
    def rounding_precision(self, value):
        if not isinstance(value, int):
            raise ValueError(
                f"Could not set rounding_precision to {value}, rounding precision must"
                " be an int"
            )
        self._rounding_precision = value

    @staticmethod
    def analyze_bulkmod(calc_path, rounding_precision):
        """
        Fit an EOS to calculate the bulk modulus from a finished bulkmod calculation
        """
        strain_paths = [path for path in calc_path.glob("strain*") if path.is_dir()]
        strain_paths = sorted(strain_paths, key=lambda d: int(d.name.split("_")[-1]))
        volumes = []
        final_energies = []
        for i, strain_path in enumerate(strain_paths):
            poscar_path = strain_path / "POSCAR"
            # search for vasprun.xml or vasprun.xml.gz
            vasprun_glob = list(strain_path.glob("vasprun.xml*"))
            if len(vasprun_glob) == 0:
                raise Exception(f"No OUTCAR available at {strain_path}")
            vasprun_path = vasprun_glob[0]
            volume = Structure.from_file(poscar_path).volume
            vasprun = Vasprun(
                filename=vasprun_path,
                parse_dos=False,
                parse_eigen=False,
                parse_potcar_file=False,
            )
            final_energy = vasprun.final_energy
            volumes.append(volume)
            final_energies.append(final_energy)
        logger.debug("Volumes")
        logger.debug(f"{volumes}")
        logger.debug("Final Energies")
        logger.debug(f"{final_energies}")
        eos_analyzer = BirchMurnaghan(volumes, final_energies)
        eos_analyzer.fit()
        bulk_modulus = np.round(eos_analyzer.b0_GPa, rounding_precision)
        logger.info(f"BULK MODULUS: {bulk_modulus}")
        b_dict = {"B": bulk_modulus}
        return b_dict

    @property
    def results(self):
        if self._results is None:
            self._results = self.analyze_bulkmod(self.calc_path, self.rounding_precision)
        return self._results



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import json
import logging
from functools import cached_property
from pathlib import Path

import numpy as np
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from vasp_manager.utils import NumpyEncoder, pgrep

logger = logging.getLogger(__name__)


class ElasticAnalyzer:
    def __init__(
        self, calc_path=None, cij=None, change_from_vasp=True, rounding_precision=3
    ):
        """
        If calc_path is specified, read cij from the OUTCAR in calc path
        If cij is specified, use it directly
        Args:
            cij (6x6 array-like[float]): stiffness tensor in Voigt notation
                -- careful! VASP does not output stiffness tensors in this notation
            calc_path (str): path to elastic calculation folder
            change_from_vasp (bool): if True, convert stiffness tensor from
                VASP's elastic tensor notation to Voigt notation
            rounding_precision (int): precision to round calculated quantities
        """
        self._cij = cij
        self._calc_path = Path(calc_path) if calc_path else calc_path
        self.change_from_vasp = change_from_vasp
        self._rounding_precision = rounding_precision
        self._results = None
        self._structure = None
        self._crystal_system = None

    @property
    def calc_path(self):
        if self._calc_path is not None:
            self.calc_path = self._calc_path
        return self._calc_path

    @calc_path.setter
    def calc_path(self, value):
        # make sure cij wasn't already defined
        # if self._cij is not None:
        #     raise Exception("Could not set calc_path as cij was already specified")
        if not value.exists():
            raise ValueError(f"Could not set calc_path to {value} as it does not exist")
        self._calc_path = value

    @property
    def structure(self):
        if self._structure is None:
            self._structure = Structure.from_file(self.calc_path / "POSCAR")
        return self._structure

    @property
    def density(self):
        return self.structure.density

    @property
    def crystal_system(self):
        if self._crystal_system is None:
            sga = SpacegroupAnalyzer(self.structure, symprec=1e-3, angle_tolerance=-1.0)
            self._crystal_system = sga.get_crystal_system()
        return self._crystal_system

    @property
    def rounding_precision(self):
        if self._rounding_precision is not None:
            self.rounding_precision = self._rounding_precision
        return self._rounding_precision

    @rounding_precision.setter
    def rounding_precision(self, value):
        if not isinstance(value, int):
            raise ValueError(
                f"Could not set rounding_precision to {value}, rounding precision must"
                " be an int"
            )
        self._rounding_precision = value

    @cached_property
    def elastic_file(self):
        return self.calc_path / "elastic_constants.txt"

    @cached_property
    def outcar_file(self):
        # search for OUTCAR or OUTCAR.gz
        outcar_glob = list(self.calc_path.glob("OUTCAR*"))
        if len(outcar_glob) == 0:
            raise Exception(f"No OUTCAR available at {self.calc_path}")
        return outcar_glob[0]

    @property
    def cij(self):
        # assumed to be in GPa
        if self._cij is None:
            self.cij = np.round(
                self._read_stiffness_tensor_file(), self.rounding_precision
            )
        else:
            self.cij = self._cij
        return self._cij

    @cij.setter
    def cij(self, values):
        if np.shape(values) != (6, 6):
            raise ValueError
        if np.asarray(values).dtype != float:
            raise ValueError
        self._cij = values

    @cached_property
    def sij(self):
        # DO NOT ROUND
        return self.get_compliance_tensor(self.cij)

    @cached_property
    def b_reuss(self):
        return np.round(self.get_B_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def b_voigt(self):
        return np.round(self.get_B_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def b_vrh(self):
        return np.round(
            self.get_VRH_average(self.b_reuss, self.b_voigt), self.rounding_precision
        )

    @cached_property
    def g_reuss(self):
        return np.round(self.get_G_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def g_voigt(self):
        return np.round(self.get_G_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def g_vrh(self):
        return np.round(
            self.get_VRH_average(self.g_reuss, self.g_voigt), self.rounding_precision
        )

    @cached_property
    def elastically_unstable(self):
        return self.check_elastically_unstable(self.cij, self.crystal_system)

    @cached_property
    def vt(self):
        return np.round(self.get_vt(self.g_vrh, self.density), self.rounding_precision)

    @cached_property
    def vl(self):
        return np.round(
            self.get_vl(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    @cached_property
    def vs(self):
        return np.round(
            self.get_vs(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    @staticmethod
    def change_elastic_constants_from_vasp(vasp_elastic_tensor):
        """
        VASP ordering of elastic constants is not the same as those expected
        by my equations below
        We should expect from Voigt notation:
        1: 11 or xx
        2: 22 or yy
        3: 33 or zz
        4: 23 or yz
        5: 13 or xz
        6: 12 or xy
        but VASP differs as it presents 1, 2, 3, 6 (xy), 4 (yz), 5 (xz)
        This function performs swapping to match expectations

        Args:
            vasp_elastic_tensor (6x6 np.array[float])
        Returns:
            elastic_tensor (6x6 np.array[float]): reordered to match Voigt notation
        """
        elastic_tensor = np.asarray(vasp_elastic_tensor).copy()
        for j in range(6):
            elastic_tensor[3, j], elastic_tensor[4, j], elastic_tensor[5, j] = (
                elastic_tensor[4, j],
                elastic_tensor[5, j],
                elastic_tensor[3, j],
            )
        for i in range(6):
            elastic_tensor[i, 3], elastic_tensor[i, 4], elastic_tensor[i, 5] = (
                elastic_tensor[i, 4],
                elastic_tensor[i, 5],
                elastic_tensor[i, 3],
            )
        return elastic_tensor

    @staticmethod
    def get_compliance_tensor(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): stiffness tensor
        Returns:
            sij (6x6 np.arrray[float]): compliance tensor
        """
        sij = np.linalg.inv(cij)
        return sij

    @staticmethod
    def get_VRH_average(mod1, mod2):
        """
        Args:
            mod1 (float):  B or G Voigt
            mode2 (float): B or G Reuss
        Returns:
            VRH_average (float)
        """
        return (mod1 + mod2) / 2.0

    @staticmethod
    def get_B_Reuss(sij):
        """
        Args:
            sij (6x6 np.arrray[float]): compliance tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        B_Reuss = 1 / (
            (sij[0, 0] + sij[1, 1] + sij[2, 2]) + 2 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
        )
        return B_Reuss

    @staticmethod
    def get_B_Voigt(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): compliance_tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        B_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2]) + 2 * (cij[0, 1] + cij[1, 2] + cij[2, 0])
        ) / 9
        return B_Voigt

    @staticmethod
    def get_G_Reuss(sij):
        """
        Args:
            sij (6x6 np.arrray[float]): compliance tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        G_Reuss = 15 / (
            4 * (sij[0, 0] + sij[1, 1] + sij[2, 2])
            - 4 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
            + 3 * (sij[3, 3] + sij[4, 4] + sij[5, 5])
        )
        return G_Reuss

    @staticmethod
    def get_G_Voigt(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): stiffness tensor
        Returns:
            G_Reuss (float): Reuss shear modulus
        """
        G_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2])
            - (cij[0, 1] + cij[1, 2] + cij[2, 0])
            + 3 * (cij[3, 3] + cij[4, 4] + cij[5, 5])
        ) / 15
        return G_Voigt

    @staticmethod
    def get_vl(bulkmod, shearmod, density):
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        return vl

    @staticmethod
    def get_vt(shearmod, density):
        vt = np.sqrt(shearmod / density)
        return vt

    @staticmethod
    def get_vs(bulkmod, shearmod, density):
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        vt = np.sqrt(shearmod / density)
        vs = (1 / 3 * (1 / vl**3 + 2 / vt**3)) ** (-1 / 3)
        return vs

    @staticmethod
    def check_elastically_unstable(cij, crystal_system):
        "returns True if compound is elastically unstable"
        eigenvalues, eigenvectors = np.linalg.eig(cij)
        born_critera_satisfied = np.all(eigenvalues > 0)

        if not born_critera_satisfied:
            logger.warning("-" * 10 + " WARNING: Elastically Unstable " + "-" * 10)
            return True

        return False

    def _make_stiffness_tensor_file(self):
        """
        Utility function that scrapes OUTCAR for elastic constants
        Writes to elastic_constants.txt
        """
        # need to get elastic dir
        elastic_table = pgrep(
            self.outcar_file,
            str_to_grep="TOTAL ELASTIC MOD",
            stop_after_first_match=True,
            after=8,
            as_string=True,
        )
        with open(self.elastic_file, "w+") as fw:
            fw.write(elastic_table)

    def _read_stiffness_tensor_file(self):
        """
        Reads vasp stiffness tensor from elastic_file

        Returns:
            elastic_tensor (6x6 np.array[float]): stiffness tensor
        """
        self._make_stiffness_tensor_file()

        with open(self.elastic_file, "r") as fr:
            raw_elastic_data = fr.readlines()
        # Skip first 3 rows as they are just header
        # Skip the first column as it contains xx, xy, etc
        elastic_data = [line.strip().split()[1:] for line in raw_elastic_data[3:]]
        # Divide by 10 to get GPa instead of kBar
        elastic_tensor = np.array(elastic_data, dtype=float) / 10.0
        if self.change_from_vasp:
            elastic_tensor = self.change_elastic_constants_from_vasp(elastic_tensor)
        return elastic_tensor

    def _analyze_elastic(self):
        """
        Grabs important quantities from the elastic calculation results

        Args:
            elastic_file (str): filepath
        Returns:
            elastic_dict (dict): dict of extracted info from
                elastic calculation
        """
        elastic_dict = {}
        elastic_dict["B_Reuss"] = self.b_reuss
        elastic_dict["B_Voigt"] = self.b_voigt
        elastic_dict["B_VRH"] = self.b_vrh
        elastic_dict["G_Reuss"] = self.g_reuss
        elastic_dict["G_Voigt"] = self.g_voigt
        elastic_dict["G_VRH"] = self.g_vrh
        elastic_dict["unstable"] = self.elastically_unstable
        elastic_dict["elastic_tensor"] = self.cij
        elastic_dict["vl"] = self.vl
        elastic_dict["vt"] = self.vt
        elastic_dict["vs"] = self.vs

        logger.debug(json.dumps(elastic_dict, cls=NumpyEncoder, indent=2))

        return elastic_dict

    @property
    def results(self):
        if self._results is None:
            self._results = self._analyze_elastic()
        return self._results



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import gzip
import json
import os
from contextlib import contextmanager

import numpy as np
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


@contextmanager
def change_directory(new_dir):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def get_pmg_structure_from_poscar(
    poscar_path,
    to_process=True,
    primitive=True,
    symprec=1e-3,
    return_sg=False,
):
    """
    Args:
        poscar_path (str)
        to_process (bool): if True, get standard reduced structure
        primitive (bool): if True, get primitive structure, else get
            conventional structure
        symprec (float): symprec for SpacegroupAnalyzer
        return_sg (bool): if True, return spacegroup number
    Returns:
        structure (pmg.Structure): structure from POSCAR
    """
    structure = Structure.from_file(poscar_path)
    if to_process:
        sga = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=-1.0)
        if primitive:
            structure = sga.get_primitive_standard_structure()
        else:
            structure = sga.get_conventional_standard_structure()
        if return_sg:
            sg = sga.get_space_group_number()
            return structure, sg
    return structure


def pcat(f_names):
    """
    Custom python-only replacement for cat

    Args:
        f_names (list): names of files to cat together
    Returns:
        catted (str)
    """
    f_contents = []
    for f_name in f_names:
        with open(f_name) as fr:
            f_content = fr.read()
        f_contents.append(f_content)
    catted = "".join(f_content for f_content in f_contents)
    return catted


def pgrep(
    f_name,
    str_to_grep,
    stop_after_first_match=False,
    after=None,
    as_string=False,
):
    """
    Custom python-only replacement for grep

    Args:
        f_name (str): path of file
        str_to_grep (str): target string
        stop_after_first_match (bool): if True, stop after first found instance of
            str_to_grep
        after (int): if not None, return {after} lines found after str_to_grep
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        matches (str | list)
    """
    opener = gzip.open if ".gz" in str(f_name) else open
    with opener(f_name, "rt") as fr:
        f_lines = [line.strip() for line in fr.readlines()]
    matches = []
    for i, line in enumerate(f_lines):
        if str_to_grep in line:
            matches.append(line)
            if after is not None:
                matches.extend(f_lines[(i + 1) : (i + after + 1)])
            if stop_after_first_match:
                break
    if as_string:
        matches = "\n".join([line for line in matches])
    return matches


def phead(f_name, n_head=1, as_string=False):
    """
    Custom python-only replacement for head

    Args:
        f_name (str): path of file
        n_tail (int): n lines to head
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        head (str | list)
    """
    opener = gzip.open if ".gz" in str(f_name) else open
    with opener(f_name, "rt") as fr:
        head = [line.strip() for line in fr.readlines()[:n_head]]
    if as_string:
        head = "\n".join([line for line in head])
    return head


def ptail(f_name, n_tail=1, as_string=False):
    """
    Custom python-only replacement for grep

    Args:
        f_name (str): path of file
        n_tail (int): n lines to tail
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        tail (str | list)
    """
    opener = gzip.open if ".gz" in str(f_name) else open
    with opener(f_name, "rt") as fr:
        tail = [line.strip() for line in fr.readlines()[-n_tail:]]
    if as_string:
        tail = "\n".join([line for line in tail])
    return tail



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import logging
from functools import cached_property

from vasp_manager.analyzer import ElasticAnalyzer
from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import pgrep, ptail
from vasp_manager.vasp_input_creator import VaspInputCreator

logger = logging.getLogger(__name__)


class ElasticCalculationManager(BaseCalculationManager):
    """
    Runs elastic deformation job workflow for a single material
    """

    def __init__(
        self,
        material_path,
        to_rerun,
        to_submit,
        primitive=True,
        ignore_personal_errors=True,
        from_scratch=False,
        tail=5,
    ):
        """
        For material_path, to_rerun, to_submit, ignore_personal_errors, and from_scratch,
        see BaseCalculationManager

        Args:
            tail (int): number of last lines to log in debugging if job failed

        Note: primitive is set to False for elastic calculations as sometimes elongated or
            distorted cells lead to erroneous results
        """
        self.tail = tail
        super().__init__(
            material_path=material_path,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done = None
        self._results = None

    @cached_property
    def mode(self):
        return "elastic"

    @cached_property
    def poscar_source_path(self):
        return self.material_path / "rlx" / "CONTCAR"

    @cached_property
    def vasp_input_creator(self):
        return VaspInputCreator(
            self.calc_path,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    def _check_use_spin(self):
        rlx_stdout = self.material_path / "rlx" / "stdout.txt"
        rlx_mags = pgrep(rlx_stdout, "mag=", stop_after_first_match=True)
        use_spin = len(rlx_mags) != 0
        return use_spin

    def setup_calc(self, increase_nodes_by_factor=2, increase_walltime_by_factor=1):
        """
        Runs elastic constants routine through VASP

        By default, requires relaxation (as the elastic constants routine needs
            the cell to be nearly at equilibrium)
        """
        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.increase_walltime_by_factor = increase_walltime_by_factor
        self.vasp_input_creator.use_spin = self._check_use_spin()
        self.vasp_input_creator.create()

        if self.to_submit:
            job_submitted = self.submit_job()
            # job status returns True if sucessfully submitted, else False
            if not job_submitted:
                self.setup_calc()

    def check_calc(self):
        """
        Checks result of elastic calculation

        Returns:
            elastic_successful (bool): if True, elastic calculation completed
                successfully
        """
        if not self.job_complete:
            logger.info(f"{self.mode.upper()} job not finished")
            return False

        stdout_path = self.calc_path / "stdout.txt"
        if not stdout_path.exists():
            # shouldn't get here unless function was called with submit=False
            logger.info(f"{self.mode.upper()} Calculation: No stdout.txt available")
            if self.to_rerun:
                self._from_scratch()
                self.setup_calc()
            return False

        vasp_errors = self._check_vasp_errors()
        if len(vasp_errors) > 0:
            all_errors_addressed = self._address_vasp_errors(vasp_errors)
            if not all_errors_addressed:
                msg = (
                    f"{self.mode.upper()} Calculation: ",
                    "Couldn't address all VASP Errors\n",
                    "\tRefusing to continue...\n",
                    f"\tVasp Errors: {vasp_errors}\n",
                )
                raise RuntimeError(msg)
            if self.to_rerun:
                logger.info(f"Rerunning {self.calc_path}")
                self._from_scratch()
                self.setup_calc()
            return False

        grep_output = pgrep(stdout_path, str_to_grep="Total")
        if len(grep_output) == 0:
            if self.to_rerun:
                logger.info(f"Rerunning {self.calc_path}")
                # calculation failed before end of first SCF cycle
                self._from_scratch()
                self.setup_calc()
            return False

        last_grep_line = grep_output[-1].strip().split()
        # last grep line looks something like 'Total: 36/ 36'
        finished_deformations = int(last_grep_line[-2].replace("/", ""))
        total_deformations = int(last_grep_line[-1])
        logger.debug(last_grep_line)
        if not finished_deformations == total_deformations:
            tail_output = ptail(stdout_path, n_tail=self.tail, as_string=True)
            logger.info(tail_output)
            logger.info(f"{self.mode.upper()} Calculation: FAILED")
            if self.to_rerun:
                logger.info(f"Rerunning {self.calc_path}")
                # increase walltime as its likely the calculation failed
                self._from_scratch()
                self.setup_calc(increase_walltime_by_factor=2)
            return False

        logger.info(f"{self.mode.upper()} Calculation: Success")
        return True

    @property
    def is_done(self):
        if self._is_done is None:
            self._is_done = self.check_calc()
        return self._is_done

    @property
    def results(self):
        if not self.is_done:
            return None
        try:
            self._results = self._analyze_elastic()
        except Exception as e:
            logger.warning(e)
            self._results = None
        return self._results

    def _analyze_elastic(self):
        """
        Gets results from elastic calculation
        """
        ea = ElasticAnalyzer(calc_path=self.calc_path)
        return ea.results



from vasp_manager.calculation_manager.bulkmod import BulkmodCalculationManager
from vasp_manager.calculation_manager.elastic import ElasticCalculationManager
from vasp_manager.calculation_manager.rlx import RlxCalculationManager
from vasp_manager.calculation_manager.rlx_coarse import RlxCoarseCalculationManager
from vasp_manager.calculation_manager.static import StaticCalculationManager



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import logging
import os
import shutil
from functools import cached_property
from pathlib import Path

import numpy as np

from vasp_manager.analyzer.bulkmod_analyzer import BulkmodAnalyzer
from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import change_directory, pgrep
from vasp_manager.vasp_input_creator import VaspInputCreator

logger = logging.getLogger(__name__)


class BulkmodCalculationManager(BaseCalculationManager):
    """
    Runs bulk modulus job workflow for a single material
    """

    def __init__(
        self,
        material_path,
        to_rerun,
        to_submit,
        primitive=True,
        ignore_personal_errors=True,
        from_relax=True,
        from_scratch=False,
        strains=None,
        tail=5,
    ):
        """
        For material_path, to_rerun, to_submit, ignore_personal_errors, and from_scratch,
        see BaseCalculationManager

        Args:
            from_relax (bool): if True, use CONTCAR from relax
            strains (array-like): optional, fractional strain along each axis for
                each deformation
                if None, use default
                default is np.linspace(start=0.925, stop=1.075, number=11)**(1/3)
                len(strains) must be odd and strains must be centered around 0
        """
        self.from_relax = from_relax
        self.strains = (
            strains
            if strains is not None
            else np.power(np.linspace(0.925, 1.075, 11), 1 / 3)
        )
        self.tail = tail
        super().__init__(
            material_path=material_path,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done = None
        self._results = None

    @cached_property
    def mode(self):
        mode_str = "bulkmod"
        if not self.from_relax:
            mode_str += "_standalone"
        return mode_str

    @cached_property
    def poscar_source_path(self):
        if self.from_relax:
            poscar_source_path = self.material_path / "rlx" / "CONTCAR"
        else:
            poscar_source_path = self.material_path / "POSCAR"
        return poscar_source_path

    @cached_property
    def vasp_input_creator(self):
        return VaspInputCreator(
            self.calc_path,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    @property
    def strains(self):
        return self._strains

    @strains.setter
    def strains(self, values):
        if np.any(values < 0.8) or np.any(values > 1.2):
            raise ValueError("Strains not in expected bounds")
        if (middle := values[int(len(values) / 2)]) != 1.0:
            raise ValueError(f"Strains not centered around 1.0: middle is {middle}")
        self._strains = values

    def _check_use_spin(self):
        # if not from_relax, use setting from calc_config
        if not self.from_relax:
            return True
        # otherwise, check if relax finished with spin
        rlx_stdout = self.material_path / "rlx" / "stdout.txt"
        rlx_mags = pgrep(rlx_stdout, "mag=", stop_after_first_match=True)
        use_spin = len(rlx_mags) != 0
        return use_spin

    def setup_calc(self, increase_nodes_by_factor=1):
        """
        Sets up an EOS bulkmod calculation
        """
        if not self.from_relax:
            msg = (
                "Running bulk modulus calculation without previous relaxation"
                "\n\t starting structure must be fairly close to equilibrium volume!"
            )
            logger.warning(msg)
            use_spin = True
        else:
            use_spin = self._check_use_spin()

        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.use_spin = use_spin
        self.vasp_input_creator.create()
        self._make_bulkmod_strains()

        if self.to_submit:
            job_submitted = self.submit_job()
            # job status returns True if sucessfully submitted, else False
            if not job_submitted:
                self.setup_calc()

    def check_calc(self):
        """
        Checks result of bulk modulus calculation

        Returns:
            bulkmod_sucessful (bool): if True, bulkmod calculation completed successfully
        """
        if not self.job_complete:
            logger.info(f"{self.mode.upper()} job not finished")
            return False

        for i, strain in enumerate(self.strains):
            middle = int(len(self.strains) / 2)
            strain_index = i - middle
            strain_name = f"strain_{strain_index}"
            strain_path = self.calc_path / strain_name
            stdout_path = strain_path / "stdout.txt"
            stderr_path = strain_path / "stderr.txt"
            if not stdout_path.exists():
                return False

            vasp_errors = self._check_vasp_errors(
                stdout_path=stdout_path, stderr_path=stderr_path
            )
            if len(vasp_errors) > 0:
                all_errors_addressed = self._address_vasp_errors(vasp_errors)
                if not all_errors_addressed:
                    msg = (
                        f"{self.mode.upper()} Calculation: ",
                        "Couldn't address all VASP Errors\n",
                        "\tRefusing to continue...\n",
                        f"\tVasp Errors: {vasp_errors}\n",
                    )
                    raise RuntimeError(msg)
                if self.to_rerun:
                    logger.info(f"Rerunning {self.calc_path}")
                    self._from_scratch()
                    self.setup_calc()
                return False

            grep_output = pgrep(stdout_path, "1 F=", stop_after_first_match=True)
            if len(grep_output) == 0:
                if self.to_rerun:
                    logger.info(f"Rerunning {self.calc_path}")
                    # increase nodes as its likely the calculation failed
                    self._from_scratch()
                    self.setup_calc(increase_nodes_by_factor=2)
                return False
        return True

    @property
    def is_done(self):
        if self._is_done is None:
            self._is_done = self.check_calc()
        return self._is_done

    @property
    def results(self):
        if not self.is_done:
            return None
        try:
            ba = BulkmodAnalyzer(calc_path=self.calc_path)
            self._results = ba.results
        except Exception as e:
            logger.warning(e)
            self._results = None
        return self._results

    def _make_bulkmod_strains(self):
        """
        Creates a set of strain directory for fitting the E-V info

        Args:
            strains (iterable of floats)
        """
        logger.info("Making strain directories")
        for i, strain in enumerate(self.strains):
            middle = int(len(self.strains) / 2)
            strain_index = i - middle
            strain_name = f"strain_{strain_index}"
            strain_path = self.calc_path / strain_name
            logger.info(strain_path)

            if not strain_path.exists():
                strain_path.mkdir()
            orig_poscar_path = self.calc_path / "POSCAR"
            strain_poscar_path = strain_path / "POSCAR"
            shutil.copy(orig_poscar_path, strain_poscar_path)

            # change second line to be {strain} rather than 1.0
            with open(strain_poscar_path, "r") as fr:
                strain_poscar = fr.read().splitlines()
            strain_poscar[1] = f"{strain}"
            logger.debug(f"{strain}")
            with open(strain_poscar_path, "w+") as fw:
                fw.write("\n".join(strain_poscar))

            with change_directory(strain_path):
                for f in ["POTCAR", "INCAR"]:
                    orig_path = Path("..") / f
                    os.symlink(orig_path, f, target_is_directory=False)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import json
import logging
import os
import pkgutil
import shutil
import warnings
from datetime import time, timedelta
from functools import cached_property
from pathlib import Path

import numpy as np
from pymatgen.io.vasp import Poscar, Potcar

from vasp_manager.utils import change_directory, get_pmg_structure_from_poscar, pcat

logger = logging.getLogger(__name__)


class VaspInputCreator:
    """
    Handles VASP file creation
    """

    def __init__(
        self,
        calc_path,
        mode,
        poscar_source_path,
        primitive=True,
        name=None,
        increase_nodes_by_factor=1,
        increase_walltime_by_factor=1,
        poscar_significant_figures=8,
        ncore_per_node_for_memory=0,
        use_spin=True,
    ):
        """
        Args:
            calc_path
            mode
            poscar_source_path
            primitive
            name
            increase_nodes_by_factor
            increase_walltime_by_factor
            poscar_significant_figures
            ncore_per_node_for_memory
            use_spin -- pass False to supress spin polarization
        """
        self.calc_path = Path(calc_path)
        self.poscar_source_path = Path(poscar_source_path)
        self.primitive = primitive
        self.increase_nodes_by_factor = int(increase_nodes_by_factor)
        self.increase_walltime_by_factor = int(increase_walltime_by_factor)
        self.name = name
        self.mode = self._get_mode(mode)
        self.poscar_significant_figures = poscar_significant_figures
        self.ncore_per_node_for_memory = ncore_per_node_for_memory
        self.use_spin = use_spin

    def _get_mode(self, mode):
        # rlx-coarse, rlx, bulkmod, stc, or elastic
        # needed to add this to ensure bulkmod or bulkmod_standalone
        # share same config
        if "bulkmod" in mode:
            mode = "bulkmod"
        return mode

    @cached_property
    def calc_config_dict(self):
        # sits in calculation folder like calculations/material_name/mode
        # first pardir is material_name/
        # second pardir is calculations/
        # config dict should sit in calculations/
        all_calcs_dir = self.calc_path.parent.parent
        fname = "calc_config.json"
        fpath = all_calcs_dir / fname
        if fpath.exists():
            with open(fpath) as fr:
                calc_config = json.load(fr)
        else:
            raise Exception(f"No {fname} found in path {all_calcs_dir.absolute()}")
        return calc_config

    @cached_property
    def computing_config_dict(self):
        all_calcs_dir = self.calc_path.parent.parent
        fname = "computing_config.json"
        fpath = all_calcs_dir / fname
        if fpath.exists():
            with open(fpath) as fr:
                computing_config = json.load(fr)
        else:
            raise Exception(f"No {fname} found in path {all_calcs_dir.absolute()}")
        return computing_config

    @cached_property
    def computer(self):
        return self.computing_config_dict["computer"]

    @cached_property
    def source_structure(self):
        num_archives = len(list(self.calc_path.glob("archive*")))
        if num_archives > 0:
            archive_name = f"archive_{num_archives-1}"
            self.poscar_source_path = self.calc_path / archive_name / "CONTCAR"
        try:
            structure = get_pmg_structure_from_poscar(
                self.poscar_source_path, primitive=self.primitive
            )
        except Exception as e:
            raise Exception(f"Cannot load POSCAR in {self.poscar_source_path}: {e}")
        return structure

    @cached_property
    def incar_template(self):
        incar_template = pkgutil.get_data(
            "vasp_manager", str(Path("static_files") / "INCAR_template")
        ).decode("utf-8")
        return incar_template

    @cached_property
    def potcar_dict(self):
        potcar_dict = json.loads(
            pkgutil.get_data(
                "vasp_manager", str(Path("static_files") / "pot_dict.json")
            ).decode("utf-8")
        )
        return potcar_dict

    @cached_property
    def q_mapper(self):
        q_mapper = json.loads(
            pkgutil.get_data(
                "vasp_manager", str(Path("static_files") / "q_handles.json")
            ).decode("utf-8")
        )
        return q_mapper

    def make_poscar(self):
        """
        Create and write a POSCAR
        """
        poscar = Poscar(self.source_structure)
        poscar_path = self.calc_path / "POSCAR"
        poscar.write_file(
            poscar_path, significant_figures=self.poscar_significant_figures
        )

    def make_potcar(self):
        """
        Create and write a POTCAR
        """
        potcar_path = self.calc_path / "POTCAR"
        potcar_dir = Path(self.computing_config_dict[self.computer]["potcar_dir"])

        el_names = [el.name for el in self.source_structure.composition]
        logger.debug(f"{self.source_structure.composition.reduced_formula}, {el_names}")
        pot_singles = [
            potcar_dir / self.potcar_dict[el_name] / "POTCAR" for el_name in el_names
        ]
        for pot_single in pot_singles:
            if not pot_single.exists():
                msg = "Unable to create POTCAR"
                msg += f"\n\t POTCAR not found at path {pot_single}"
                raise Exception(msg)

        potcar = pcat(pot_singles)
        with open(potcar_path, "w+") as fw:
            fw.write(potcar)

    @property
    def n_nodes(self):
        # start with 1 node per 32 atoms
        num_nodes = (len(self.source_structure) // 32) + 1
        if self.computer == "quest":
            # quest has 4x smaller nodes than perlmutter
            num_nodes *= 4
        num_nodes *= self.increase_nodes_by_factor
        return num_nodes

    @property
    def n_procs(self):
        # typically request all processors on each node, and then
        # leave some ~4/node empty for memory
        n_procs = (
            self.n_nodes * self.computing_config_dict[self.computer]["ncore_per_node"]
        )
        return n_procs

    @property
    def n_procs_used(self):
        ncore_per_node = self.computing_config_dict[self.computer]["ncore_per_node"]
        if self.mode == "elastic":
            if self.computer == "quest":
                self.ncore_per_node_for_memory += 8
        return self.n_nodes * (ncore_per_node - self.ncore_per_node_for_memory)

    @cached_property
    def d_f_block(self):
        d_f_block = json.loads(
            pkgutil.get_data(
                "vasp_manager", str(Path("static_files") / "d_f_block.json")
            ).decode("utf-8")
        )
        return d_f_block

    def _check_needs_spin_polarization(self, composition_dict):
        needs_spin_polarization = False
        for el_name in composition_dict:
            if (
                el_name in self.d_f_block["d_block"]
                or el_name in self.d_f_block["f_block"]
            ):
                needs_spin_polarization = True
        return needs_spin_polarization

    def _get_auto_magmom(self, composition_dict):
        magmom_string = "MAGMOM ="
        for el_name in composition_dict:
            if el_name in self.d_f_block["d_block"]:
                init_mag = "5.0"
            elif el_name in self.d_f_block["f_block"]:
                init_mag = "7.0"
            else:
                init_mag = "0.0"
            magmom_atom = str(int(composition_dict[el_name])) + "*" + init_mag
            magmom_string += f" {magmom_atom}"
        return magmom_string

    @cached_property
    def hubbards(self):
        hubbards = json.loads(
            pkgutil.get_data(
                "vasp_manager", os.path.join("static_files", "hubbards.json")
            ).decode("utf-8")
        )
        return hubbards

    def _check_needs_dftu(self, hubbards_type, composition_dict):
        if not hubbards_type:
            return False

        needs_dftu = False
        for el_name in composition_dict:
            if el_name in self.hubbards[hubbards_type]:
                needs_dftu = True
        return needs_dftu

    def _get_ldau_string(self, hubbards_type, composition_dict):
        u_string = "LDAUU ="
        j_string = "LDAUJ ="
        l_string = "LDAUL ="
        hubbards = self.hubbards[hubbards_type]

        for el_name in composition_dict:
            if el_name in hubbards:
                u_val = hubbards[el_name]["U"]
                l_val = hubbards[el_name]["L"]
            else:
                u_val = 0.0
                l_val = -1
            u_string += f" {float(u_val)}"
            j_string += " 0.0"
            l_string += f" {l_val}"

        ldau_string = "LDAU = .TRUE.\n"
        ldau_string += "LDAUPRINT = 1\n"
        ldau_string += f"{u_string}\n"
        ldau_string += f"{j_string}\n"
        ldau_string += f"{l_string}"
        return ldau_string

    def _get_lmaxmix(self, composition_dict):
        d_block_elements = self.d_f_block["d_block"]
        f_block_elements = self.d_f_block["f_block"]
        lmaxmix = 2
        for element in composition_dict:
            if element in d_block_elements:
                lmaxmix = 4
        for element in composition_dict:
            if element in f_block_elements:
                lmaxmix = 6
        return lmaxmix

    def make_incar(self):
        """
        Create and write an INCAR

        Need to modify this to account for spin/magmom
        Current kpoints coming from the kspacing tag in the INCAR,
            but future versions should include ability to make kpoints from kppra
        """
        incar_path = self.calc_path / "INCAR"
        ncore = self.computing_config_dict[self.computer]["ncore"]
        calc_config = self.calc_config_dict[self.mode]

        if calc_config["ispin"] not in [1, "auto"]:
            raise RuntimeError("ISPIN must be set to 1 or auto")

        if calc_config["hubbards"] not in [None, False, "wang"]:
            raise RuntimeError("hubbards must be set to None, False, or 'wang'")
        if calc_config["hubbards"] == "wang" and calc_config["gga"] != "PE":
            raise RuntimeError("DFT+U is only support for PBE")

        if calc_config["iopt"] != 0 and calc_config["potim"] != 0:
            raise RuntimeError("To use IOPT != 0, POTIM must be set to 0")

        composition_dict = self.source_structure.composition.as_dict()
        # read POTCAR
        potcar_path = self.calc_path / "POTCAR"
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            potcar = Potcar.from_file(potcar_path)
        n_electrons = 0
        for potcar_single in potcar:
            n_electrons += (
                potcar_single.nelectrons * composition_dict[potcar_single.element]
            )
        # make n_bands divisible by NCORE (VASP INCAR tag)
        calc_config["ncore"] = ncore
        calc_config["nbands"] = int(np.ceil(0.75 * n_electrons / ncore) * ncore)

        needs_spin_polarization = self._check_needs_spin_polarization(composition_dict)
        use_spin_polarization = (
            needs_spin_polarization and calc_config["ispin"] == "auto" and self.use_spin
        )
        if use_spin_polarization:
            ispin = 2
            magmom_line = self._get_auto_magmom(composition_dict)
        else:
            ispin = 1

        needs_dftu = self._check_needs_dftu(calc_config["hubbards"], composition_dict)
        use_dftu = needs_dftu and calc_config["hubbards"]
        if use_dftu:
            lmaxmix = self._get_lmaxmix(composition_dict)
            lmaxmix_line = f"LMAXMIX = {lmaxmix}"
            ldau_string = self._get_ldau_string(calc_config["hubbards"], composition_dict)

        # Add lines to the vaspq file
        incar_tmp = self.incar_template.split("\n")
        for i, line in enumerate(incar_tmp):
            # add extra flags for spin polarization
            if "ISPIN" in line:
                incar_tmp[i] = f"ISPIN = {ispin}"
                if use_spin_polarization:
                    incar_tmp.insert(i + 1, magmom_line)
            if "LVTOT" in line and use_spin_polarization:
                lorbit_line = "LORBIT = 11"
                incar_tmp.insert(i + 1, lorbit_line)
            # add extra flags for DFT+U
            if "DFT+U" in line and use_dftu:
                incar_tmp.insert(i + 1, lmaxmix_line)
                incar_tmp.insert(i + 1, ldau_string)
            # add extra flags for elastic mode
            if self.mode == "elastic":
                if "KSPACING" in line:
                    nfree_line = "NFREE = {nfree}"
                    symprec_line = "SYMPREC = {symprec}"
                    incar_tmp.insert(i + 1, symprec_line)
                    incar_tmp.insert(i + 1, nfree_line)
                if "NCORE" in line:
                    # elastic calculation won't run unless NCORE=1
                    incar_tmp[i] = "NCORE = 1"
        incar_tmp = "\n".join([line for line in incar_tmp])
        incar = incar_tmp.format(**calc_config)
        logger.debug(incar)
        with open(incar_path, "w+") as fw:
            fw.write(incar)

    def make_vaspq(self):
        """
        Create and write vasp.q file
        """
        vaspq_path = self.calc_path / "vasp.q"
        calc_config = self.calc_config_dict[self.mode]

        # create pad string for job naming to differentiate in the queue
        match self.mode:
            case "rlx-coarse" | "rlx":
                if self.mode == "rlx":
                    pad_string = "r"
                elif self.mode == "rlx-coarse":
                    pad_string = "rc"
                mode = "rlx"
            case "static":
                pad_string = "s"
                mode = "static"
            case "bulkmod":
                pad_string = "b"
                mode = "bulkmod"
            case "elastic":
                pad_string = "e"
                mode = "elastic"
            case _:
                raise ValueError(
                    "Calculation type {self.mode} not in supported calculation types"
                    "of VaspInputCreator"
                )

        if self.name is None:
            jobname = pad_string + self.source_structure.composition.reduced_formula
        else:
            jobname = pad_string + self.name

        # convert walltime into seconds for increase_walltime_by_factor
        walltime_iso = time.fromisoformat(calc_config["walltime"])
        # walltime_duration is in seconds
        walltime_duration = timedelta(
            hours=walltime_iso.hour,
            minutes=walltime_iso.minute,
            seconds=walltime_iso.second,
        )
        walltime_duration *= self.increase_walltime_by_factor
        # convert to HH:MM:SS
        walltime = str(walltime_duration)
        # cut walltime short by 1 minute so job metrics log properly
        timeout = walltime_duration.seconds - 60
        # quest uses mpirun which needs timeout in seconds
        # otherwise, convert it back to HH:MM:SS
        if not self.computer == "quest":
            timeout = str(timedelta(seconds=timeout))

        computer_config = self.computing_config_dict[self.computer].copy()
        ncore_per_node = self.n_procs_used // self.n_nodes
        computer_config.update(
            {
                "n_nodes": self.n_nodes,
                "n_procs": self.n_procs,
                "ncore_per_node": ncore_per_node,
                "jobname": jobname,
                "walltime": walltime,
                "timeout": timeout,
            }
        )

        q_name = self.q_mapper[self.computer][mode]
        vaspq_tmp = pkgutil.get_data(
            "vasp_manager", str(Path("static_files") / q_name)
        ).decode("utf-8")
        vaspq = vaspq_tmp.format(**computer_config)
        logger.debug(vaspq)
        with open(vaspq_path, "w+") as fw:
            fw.write(vaspq)

    def make_archive_and_repopulate(self):
        """
        Make an archive of a VASP calculation and copy back over relevant files
        """
        with change_directory(self.calc_path):
            contcar_path = Path("CONTCAR")
            contcar_exists = contcar_path.exists()
            if contcar_exists:
                contcar_is_empty = contcar_path.stat().st_size == 0
            else:
                contcar_is_empty = True

            if contcar_is_empty:
                # if CONTCAR is empty, don't make an archive and clean up
                all_files = [f for f in Path(".").glob("*") if f.is_file()]
                for f in all_files:
                    os.remove(f)
            else:
                # make the archive
                num_previous_archives = len(list(Path(".").glob("archive*")))
                archive_name = Path(f"archive_{num_previous_archives}")
                logger.info(f"Making {archive_name}...")
                archive_name.mkdir()

                all_files = [f for f in Path(".").glob("*") if f.is_file()]
                for f in all_files:
                    shutil.move(f, archive_name)

        self.create()

    def create(self):
        """
        Make VASP input files

        Don't touch the order! make_incar and make_vaspq rely on the poscar and
        potcar existing already
        """
        if not self.calc_path.exists():
            self.calc_path.mkdir()
        self.make_poscar()
        self.make_potcar()
        self.make_incar()
        self.make_vaspq()



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import logging
from functools import cached_property

import numpy as np

from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import get_pmg_structure_from_poscar, pgrep, ptail
from vasp_manager.vasp_input_creator import VaspInputCreator

logger = logging.getLogger(__name__)


class RlxCalculationManager(BaseCalculationManager):
    """
    Runs relaxation job workflow for a single material
    """

    def __init__(
        self,
        material_path,
        to_rerun,
        to_submit,
        primitive=True,
        ignore_personal_errors=True,
        from_coarse_relax=True,
        from_scratch=False,
        tail=5,
        max_reruns=3,
        magmom_per_atom_cutoff=0.0,
    ):
        """
        For material_path, to_rerun, to_submit, ignore_personal_errors, and from_scratch,
        see BaseCalculationManager

        Args:
            from_coarse_relax (bool): if True, use CONTCAR from coarse relax
            tail (int): number of last lines to log in debugging if job failed
        """
        self.from_coarse_relax = from_coarse_relax
        self.tail = tail
        self.max_reruns = max_reruns
        self.magmom_per_atom_cutoff = magmom_per_atom_cutoff
        super().__init__(
            material_path=material_path,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done = None
        self._results = None

    @cached_property
    def mode(self):
        return "rlx"

    @cached_property
    def poscar_source_path(self):
        if self.from_coarse_relax:
            poscar_source_path = self.material_path / "rlx-coarse" / "CONTCAR"
        else:
            poscar_source_path = self.material_path / "POSCAR"
        return poscar_source_path

    @cached_property
    def vasp_input_creator(self):
        return VaspInputCreator(
            self.calc_path,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    def setup_calc(self, increase_nodes_by_factor=1, make_archive=False, use_spin=True):
        """
        Sets up a fine relaxation
        """
        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.use_spin = use_spin

        if make_archive:
            self.vasp_input_creator.make_archive_and_repopulate()
        else:
            self.vasp_input_creator.create()

        if self.to_submit:
            job_submitted = self.submit_job()
            if not job_submitted:
                self.setup_calc()

    def check_calc(self):
        """
        Checks if calculation has finished and reached required accuracy

        Returns:
            relaxation_successful (bool): if True, relaxation completed successfully
        """
        if not self.job_complete:
            logger.info(f"{self.mode.upper()} not finished")
            return False

        stdout_path = self.calc_path / "stdout.txt"
        if not stdout_path.exists():
            # calculation never actually ran
            # shouldn't get here unless function was called with submit=False
            # or job was manually cancelled
            logger.info(f"{self.mode.upper()} Calculation: No stdout.txt available")
            if self.to_rerun:
                self._cancel_previous_job()
                self.setup_calc()
            return False

        vasp_errors = self._check_vasp_errors()
        if len(vasp_errors) > 0:
            all_errors_addressed = self._address_vasp_errors(vasp_errors)
            if not all_errors_addressed:
                msg = (
                    f"{self.mode.upper()} Calculation: ",
                    "Couldn't address all VASP Errors\n",
                    "\tRefusing to continue...\n",
                    f"\tVasp Errors: {vasp_errors}\n",
                )
                raise RuntimeError(msg)
            if self.to_rerun:
                logger.info(f"Rerunning {self.calc_path}")
                self.setup_calc(make_archive=True)
            return False

        previous_magmom_per_atom = self._parse_magmom_per_atom()
        if previous_magmom_per_atom is None:
            use_spin = False
        else:
            use_spin = np.abs(previous_magmom_per_atom) >= self.magmom_per_atom_cutoff

        tail_output = ptail(stdout_path, n_tail=self.tail, as_string=True)
        grep_output = pgrep(
            stdout_path, "reached required accuracy", stop_after_first_match=True
        )
        if len(grep_output) == 0:
            archive_dirs = list(self.calc_path.glob("archive*"))
            if len(archive_dirs) >= self.max_reruns - 1:
                msg = (
                    "Many archives exist, calculations may not be converging\n"
                    "\t Refusing to continue..."
                )
                logger.error(msg)
                return False

            logger.warning(f"{self.mode.upper()} FAILED")
            logger.debug(tail_output)
            if self.to_rerun:
                logger.info(f"Rerunning {self.calc_path}")
                # increase nodes as its likely the calculation failed
                self.setup_calc(
                    increase_nodes_by_factor=2, make_archive=True, use_spin=use_spin
                )
            return False

        # in the case that the calculation finishes but results in a spin value lower
        # than the cutoff, rerun it without spin
        if not use_spin and previous_magmom_per_atom is not None and self.to_rerun:
            logger.info(f"Rerunning {self.calc_path}")
            self.setup_calc(make_archive=True, use_spin=False)
            return False

        logger.info(f"{self.mode.upper()} Calculation: reached required accuracy")
        logger.debug(tail_output)
        return True

    def check_volume_difference(self):
        """
        Checks relaxation runs for volume difference

        if abs(volume difference) is >= 5%, reruns relaxation
        only checks for mode='rlx' as that's the structure for further calculation

        Returns:
            volume_converged (bool): if True, relaxation completed successfully
        """
        original_poscar_path = self.material_path / "POSCAR"
        poscar_path = self.calc_path / "POSCAR"
        contcar_path = self.calc_path / "CONTCAR"
        try:
            orig_structure, orig_spacegroup = get_pmg_structure_from_poscar(
                original_poscar_path, return_sg=True
            )
            p_structure, p_spacegroup = get_pmg_structure_from_poscar(
                poscar_path, return_sg=True
            )
            c_structure, c_spacegroup = get_pmg_structure_from_poscar(
                contcar_path, return_sg=True
            )
        except Exception as e:
            logger.error(f"  RLX CONTCAR doesn't exist or is empty: {e}")
            return False

        if orig_spacegroup == c_spacegroup:
            logger.debug(
                f"  Spacegroups match orig-{orig_spacegroup} == rlx-{c_spacegroup}"
            )
        else:
            logger.warning(
                "   Warning: spacegroups do not match "
                + f"orig-{orig_spacegroup} != rlx-{c_spacegroup}"
            )

        volume_diff = (c_structure.volume - p_structure.volume) / p_structure.volume
        if np.abs(volume_diff) >= 0.05:
            logger.warning(f"  NEED TO RE-RELAX: dV = {volume_diff:.4f}")
            volume_converged = False
            previous_magmom_per_atom = self._parse_magmom_per_atom()
            if previous_magmom_per_atom is None:
                use_spin = False
            else:
                use_spin = np.abs(previous_magmom_per_atom) >= self.magmom_per_atom_cutoff
            if self.to_rerun:
                self.setup_calc(make_archive=True, use_spin=use_spin)
        else:
            logger.info("  RLX volume converged")
            logger.debug(f"  dV = {volume_diff:.4f}")
            volume_converged = True
        orig_volume_diff = (
            c_structure.volume - orig_structure.volume
        ) / orig_structure.volume
        self._results = {}
        self._results["initial_spacegroup"] = orig_spacegroup
        self._results["relaxed_spacegroup"] = c_spacegroup
        self._results["total_dV"] = np.round(orig_volume_diff, 4)
        return volume_converged

    @property
    def is_done(self):
        if self._is_done is None:
            self._is_done = False
            calc_done = self.check_calc()
            if calc_done:
                volume_converged = self.check_volume_difference()
                if volume_converged:
                    self._is_done = True
        return self._is_done

    @property
    def results(self):
        if not self.is_done:
            self._results = None
        return self._results



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import os
import shutil
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path

import numpy as np

from vasp_manager.job_manager import JobManager
from vasp_manager.utils import get_pmg_structure_from_poscar, pgrep


class BaseCalculationManager(ABC):
    """
    Runs vasp job workflow for a single material
    """

    def __init__(
        self,
        material_path,
        to_rerun,
        to_submit,
        primitive=True,
        ignore_personal_errors=True,
        from_scratch=False,  # DANGEROUS, WILL DELETE PREVIOUS CALCULATION
    ):
        """
        Args:
            material_path (str): path for a single material
                ex. calculations/AlAs
            to_rerun (bool): if True, rerun failed calculations
            to_submit (bool): if True, submit calculations to job manager
            primitive (bool): if True, find primitive cell, else find conventional cell
            ignore_personal_errors (bool): if True, ignore job submission errors
                if on personal computer
            from_scratch (bool): if True, remove the calculation's folder and
                restart
                note: DANGEROUS
        """
        self.material_path = Path(material_path)
        self.to_rerun = to_rerun
        self.to_submit = to_submit
        self.primitive = primitive
        self.job_manager = JobManager(
            calc_path=self.calc_path, ignore_personal_errors=ignore_personal_errors
        )

        self.from_scratch = from_scratch
        if from_scratch:
            self._from_scratch()

    @property
    @abstractmethod
    def mode(self):
        pass

    @property
    @abstractmethod
    def is_done(self):
        pass

    @property
    @abstractmethod
    def vasp_input_creator(self):
        pass

    @cached_property
    def calc_path(self):
        return self.material_path / self.mode

    @cached_property
    def material_name(self):
        return self.material_path.name

    @abstractmethod
    def poscar_source_path(self):
        pass

    @abstractmethod
    def setup_calc(self):
        pass

    @abstractmethod
    def check_calc(self):
        pass

    @property
    def job_exists(self):
        return self.job_manager.job_exists

    @property
    def job_complete(self):
        return self.job_manager.job_complete

    @property
    def stopped(self):
        return (self.material_path / "STOP").exists()

    def submit_job(self):
        return self.job_manager.submit_job()

    def _cancel_previous_job(self):
        jobid_path = self.calc_path / "jobid"
        if jobid_path.exists():
            with open(jobid_path) as fr:
                jobid = fr.read().strip()
            cancel_job_call = f"scancel {jobid}"
            os.system(cancel_job_call)
            os.remove(jobid_path)

    def _from_scratch(self):
        self._cancel_previous_job()
        shutil.rmtree(self.calc_path)

    def _parse_magmom(self):
        stdout_path = self.calc_path / "stdout.txt"
        mag_lines = pgrep(stdout_path, "mag=")
        # if "mag=" not found in stdout, set magmom=None
        if len(mag_lines) == 0:
            magmom = None
        else:
            total_mag = mag_lines[-1].split()[-1]
            magmom = float(total_mag)
        return magmom

    def _parse_magmom_per_atom(self):
        total_magmom = self._parse_magmom()
        if total_magmom is None:
            return None
        structure = get_pmg_structure_from_poscar(self.poscar_source_path)
        magmom_per_atom = total_magmom / len(structure)
        return magmom_per_atom

    def _check_vasp_errors(self, stdout_path=None, stderr_path=None):
        """
        Find VASP errors in stdout and stderr
        """
        if stdout_path is None:
            stdout_path = self.calc_path / "stdout.txt"
        if stderr_path is None:
            stderr_path = self.calc_path / "stderr.txt"
        errors = set()

        with open(stdout_path) as fr:
            stdout = fr.readlines()
        for line in stdout:
            if "Sub-Space-Matrix" in line:
                errors.add("Sub-Space-Matrix")

        with open(stderr_path) as fr:
            stderr = fr.readlines()
        for line in stderr:
            if "oom-kill" in line:
                errors.add("OOM")

        return errors

    def _address_vasp_errors(self, errors):
        """
        Args:
            errors (set): set of errors found in stdout or stderr
        Returns:
            all_errors_addressed (bool): if True, all errors could
                be fixed automatically. If False, some errors
                could not be handled
        """
        vic = self.vasp_input_creator

        errors_addressed = {e: False for e in errors}
        for error in errors:
            match error:
                case "Sub-Space-Matrix":
                    vic.calc_config_dict[self.mode]["algo"] = "Fast"
                    errors_addressed[error] = True
                case "OOM":
                    if vic.computer == "quest":
                        ncore_per_node_for_memory = 8
                    else:
                        ncore_per_node_for_memory = 32
                    vic.ncore_per_node_for_memory = ncore_per_node_for_memory
                    errors_addressed[error] = True
                case _:
                    errors_addressed[error] = False
        all_errors_addressed = np.all([v for v in errors_addressed.values()])
        return all_errors_addressed



from setuptools import setup

with open("README.md") as file:
    long_description = file.read()

setup(
    name="vasp_manager",
    version="1.0.4",
    description="A simple package to run and analyze VASP calculations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["vasp_manager"],
    url="https://github.com/dgaines2/vasp_manager",
    author="Dale Gaines II",
    author_email="dalegainesii@gmail.com",
    license="MIT",
    include_package_data=True,
    python_requires=">=3.10.0",
    install_requires=[
        "numpy>=1.22",
        "pandas>=1.4",
        "pymatgen>=2022.5",
    ],
    extras_require={"dev": ["black", "isort", "pre-commit", "pytest"]},
)
